/* This program solves the 0-1 Knapsack problem using Branch and Bound
 * See "Knapsack Problems. Algorithms and Computer Implementations." by
 * Silvano Martello and Paolo Toth. John Wiley and Sons.
 * This is approach focuses in the aspects of general
 * parallel branch and bound and not in the concrete features of knapsack
 * problems.
 * Author: Casiano Rodriguez Leon. Email: casiano@ull.es
*/
#define FORK95

#ifdef FORK95
#include <fork.h>
#include <io.h>
#else
#include <stdio.h>
#endif
#include <stdlib.h>

#define MAX 4096
#define STRIDE 1
#define  INFINITY 8192

typedef int number;
#define ns "%d\n"

number N, M;
number w[MAX], p[MAX];
number bestSol = -INFINITY;

FILE *in;

void readKnap(void) {
  number i;

  if ((in = fopen("knap.dat","r+")) == NULL) 
    fprintf(stderr, "Cannot open input file.\n");
  fscanf(in,ns,&N);
  fscanf(in,ns,&M);
  printf("N = "); printf(ns,N);
  printf("M = "); printf(ns,M);
  for(i=0;i<N;i++) {
    fscanf(in,ns,&w[i]);
    fscanf(in,ns,&p[i]);
    if (i%STRIDE == 0) {
      printf("w = "); printf(ns,w[i]);
      printf("p = "); printf(ns,p[i]);
    }  
  }
  w[N] = INFINITY; p[N] = 0; /* sentinel object */
  fclose(in);
}

void lowerUpper(number k, number C, number P,number *L, number *U) {
  /* naive approach */
  number i, weig, prof;

  if (C < 0) {
	 *L = -INFINITY; *U = -INFINITY;
  }
  else {
	 for(i=k, weig = 0, prof = P; weig <= C; i++) {
		weig += w[i]; prof += p[i];
	 }
	 /* invariant:  weig > C */
	 i--;
	 weig -= w[i];
	 prof -= p[i];
	 *L = prof;  /* see Martello and Toth pag 16. Th 2.1 */
	 *U = prof+(p[i]*(C-weig))/w[i];
  }
}

number knap(number k, number C, number P) {
  number L, U;

  if (k < N) {
	 lowerUpper(k,C,P,&L,&U);
	 if (bestSol < L) bestSol = L;
	 /* L <= bestSol <= U */
	 if (bestSol < U) {
		int next;

      next = k+1;
		knap(next,C-w[k], P+p[k]);
      knap(next,C, P);
    }
  }
  return bestSol;
}

int main(void) {

  readKnap();
  printf("\nsol = ");
  printf(ns, knap(0 /* next object */,
		  M /* remaining capacity */,
		  0 /* profit */)
         );
  return 1;
}
